(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: plasma_inodecache.ml 283 2010-10-28 00:19:50Z gerd $ *)

(* TODO Improve:
    - config for ic_size (and maybe other params like n_fill and n_evict)
    - n_fill could be a value that is changed dynamically. If the file
      is rarely modified, n_fill should be big. If it is frequently modified
      small n_fill is better.
    - check_up_to_date: Maybe this optimization helps sometimes: When
      a number of blocks are read from the same inode, and this happens
      concurrently, the up-to-date checks could be merged. This might be
      difficult to organize, though - if more up-to-date checks are expected
      already requested checks should be minimally delayed to increase the
      chance that such a merge can happen. So the check_up_to_date
      function should have another parameter [p] which is the estimate of
      soon coming calls. Only if this parameter is > 0 the delay happens.
      There is an upper bound for the delay which is a function of [f]
      (experiments are needed to get the exact formula).

 *)

open Plasma_rpcapi_aux
open Plasma_util.Operators
open Printf

module Fsys = Plasma_rpcapi_clnt.Make'Filesystem(Rpc_proxy.ManagedClient)
module Ic = Plasma_rpcapi_clnt.Make'Inode_cache(Rpc_proxy.ManagedClient)

(** [ic] contains inode information for a cached inode *)
type ic =
    { cache_inode : int64;
      mutable cache_val_ts : float; (* when cache was last validated *)
      mutable cache_tmo_ts : float; (* when cache must time out *)
      mutable cache_ii : inodeinfo;
      mutable cache_blocks : (int64, blockinfo list) Hashtbl.t
	(** [cache_blocks] is an incomplete mapping of indexes to
	    blockinfos
	 *)
    }


module IC = struct
  type t = int64   (* inode number *)
  let compare = Int64.compare
end

module ICLRU = Plasma_util.LRU(IC)

type ic_manager =
    { mutable cacheclient : Rpc_proxy.ManagedClient.t option;
      (** If none, there is no inodecache functionality. 
	  Otherwise this is the RPC client *)

      inodecache : (int64, ic) Hashtbl.t;
      (** Cache per inode *)

      inodecache_lru : ICLRU.t;
      (** For determining the oldest cache member *)

      inodecache_refresh_e : (int64, rinodeinfo Uq_engines.engine) Hashtbl.t;
      (** For each inode this is an engine loading a fresh [ii]. Avoids
	  that several queries are emitted for the same inode at the same time
       *)

      inodecache_getting_blocks : 
	(int64, (int64 * int * unit Uq_engines.engine)) Hashtbl.t;
      (** Whether there is a get_blocks query for this inode *)

      mutable ic_size : int;
      (** max number of cached inodes *)

      esys : Unixqueue.event_system
    }


class type lazy_transaction = object
  method get_tid_e : unit -> int64 Uq_engines.engine
  method release_e : unit -> unit Uq_engines.engine
end


let create_manager esys =
  { cacheclient = None;
    inodecache = Hashtbl.create 128;
    inodecache_lru = ICLRU.create();
    inodecache_getting_blocks = Hashtbl.create 7;
    inodecache_refresh_e = Hashtbl.create 7;
    ic_size = 1024;
    esys = esys;
  }


let enable c mc =
  c.cacheclient <- Some mc

let disable c =
  match c.cacheclient with
    | None -> ()
    | Some cc ->
	Rpc_proxy.ManagedClient.trigger_shutdown
	  cc (fun () -> ());
	c.cacheclient <- None 


(** The inodecache stores the inodeinfo record, plus a subset of the
    blockinfo records. Every access to the inodecache is validated -
    it is checked whether inodeinfo is the committed version. If so,
    the cached values can be taken.

    This assumes that the validating is_up_to_date call is very cheap.
    This is achieved by separating the inodecache server into its own
    process, and avoiding db lookups most of the time.

    There is also a cache for blocklists. We must not cache these for 
    too long, however, because these lists include the [node_alive]
    flag.
 *)

let ic_evict c =
  (** Limit to ic_size entries *)
  while Hashtbl.length c.inodecache > c.ic_size do
    let oldest = ICLRU.take c.inodecache_lru in
    Hashtbl.remove c.inodecache oldest
  done;
  ()


let ic_update c inode ii =
  (** Update the cache with ii if possible *)
  Netlog.logf `Debug "ic_update: inode=%Ld ii.seqno=%Ld ii.committed=%B"
    inode ii.seqno ii.committed;
  if ii.committed && c.cacheclient <> None then (
    let old_seqno, exists =
      try (Hashtbl.find c.inodecache inode).cache_ii.seqno, true
      with Not_found -> (-1L), false in
    if ii.seqno > old_seqno then (
      (* There is a newer version. Replace the old one. *)
      let now = Unix.gettimeofday() in
      let ce =
	{ cache_inode = inode;
	  cache_val_ts = now;
	  cache_tmo_ts = now +. 60.0;     (* FIXME *)
	  cache_ii = ii;
	  cache_blocks = Hashtbl.create 3;
	} in
      Hashtbl.replace c.inodecache inode ce;
      ICLRU.add inode c.inodecache_lru;
      Netlog.logf `Debug "ic_update: inode=%Ld updating cache" inode;
      if not exists then
	ic_evict c
    )
    else
      if exists then (
	(* The version in the cache is already as good or newer. If it is
	   as good, we can at least update the validation timestamp.
	 *)
	if ii.seqno = old_seqno then (
	  let now = Unix.gettimeofday() in
	  let old_ce = Hashtbl.find c.inodecache inode in
	  old_ce.cache_val_ts <- now;
	  old_ce.cache_tmo_ts <- now +. 60.0;     (* FIXME *)
	  Netlog.logf `Debug "ic_update: inode=%Ld confirming cache" inode;
	);
	(* Anyway, update the position in the LRU list. *)
	ICLRU.add inode c.inodecache_lru  (* "touch" *)
      )
  )


let ic_touch c inode =
  (** "touches" the cache entry so that it becomes the most recently
      accessed entry
   *)
  if ICLRU.mem inode c.inodecache_lru then
    ICLRU.add inode c.inodecache_lru;
  ()


let get_ii_e c mc tid inode =
  (** Get inodeinfo directly from namenode *)
  Netlog.logf `Debug "get_ii_e(%Ld)" inode;
  Plasma_util.rpc_engine mc Fsys.V1.get_inodeinfo'async (tid,inode)
  ++ (fun r ->
	( match r with
	    | `ok ii -> ic_update c inode ii
	    | _ -> ()
	);
	eps_e (`Done r) c.esys
     )

let get_ii_serialized_e c mc tid inode fresh =
  (** Same, but serialized, so that usually at most one [get_ii_e] is active
      at a time for the same inode. By setting [fresh] one can enforce that
      a new [get_ii_e] is started (possibly in parallel to an already running
      one)
   *)
  try
    if fresh then raise Not_found;
    let e = Hashtbl.find c.inodecache_refresh_e inode in
    Netlog.logf `Debug "get_ii_serialized_e(%Ld): already in progress" inode;
    e
  with
    | Not_found ->
	Netlog.logf `Debug "get_ii_serialized_e(%Ld): starting" inode;
	let e = get_ii_e c mc tid inode in
	Hashtbl.replace c.inodecache_refresh_e inode e;
	let drop _ = Hashtbl.remove c.inodecache_refresh_e inode in
	Uq_engines.when_state
	  ~is_done:drop ~is_error:drop ~is_aborted:drop e;
	e


let invalidate_ii c inode =
  (* invalidates the current version in the cache, but does not prevent
     the ongoing refresh
   *)
  Hashtbl.remove c.inodecache inode;
  ICLRU.remove inode c.inodecache_lru


let get_ii_from_cache_e c inode =
  (** Returns inodeinfo from cache (after validation) *)
  try
    let ce = Hashtbl.find c.inodecache inode in
    ( match c.cacheclient with
	| None ->
	    raise Not_found
	| Some ic ->
	    (** we do not validate each time, but based on a time criterion
	     *)
	    let now = Unix.gettimeofday() in
	    let way_too_old = now >= ce.cache_tmo_ts in
	       (* after a certain timeout we never take the cache *)
	    let gratuitously_valid = now < ce.cache_val_ts +. 1.0 in
	       (* shortly after a validation we believe the cache *)
	    ( if gratuitously_valid then
		(eps_e (`Done true) c.esys)
	      else 
		if way_too_old then
		  (eps_e (`Done false) c.esys)
		else (
		  Netlog.logf `Debug
		    "get_ii_from_cache_e: validating inode=%Ld" inode;
		  Plasma_util.rpc_engine
		    ic Ic.V1.is_up_to_date'async (inode,ce.cache_ii)
		)
	    )
	    ++ (fun flag ->
		  Netlog.logf `Debug "get_ii_from_cache_e(%Ld): valid=%B"
		      inode flag;
		  if flag then (
		    ic_touch c inode;
		    if not gratuitously_valid then
		      ce.cache_val_ts <- now;
		    eps_e (`Done (Some ce.cache_ii)) c.esys
		  )
		  else(
		    invalidate_ii c inode;
		    eps_e (`Done None) c.esys
		  )
	       )
    )
  with
    | Not_found ->
	Netlog.logf `Debug "get_ii_from_cache_e(%Ld): not found" inode;
	eps_e (`Done None) c.esys


let get_ii_via_cache_e c mc lt inode =
  (** Returns inodeinfo from cache (after validation), or loads it *)
  get_ii_from_cache_e c inode
  ++ (function
	| Some ii ->
	    eps_e (`Done(`ok ii)) c.esys
	| None ->
	    lt # get_tid_e() ++
	      (fun tid ->
		 get_ii_serialized_e c mc tid inode false)
     )


let check_up_to_date_e c mc lt inode seqno =
  (** Returns whether this seqno is the seqno of the inode *)
  ( match c.cacheclient with
      | None -> 
	  Netlog.logf `Debug "check_up_to_date_e(%Ld): inactive" inode;
	  eps_e (`Done false) c.esys
      | Some ic ->
	  Netlog.logf `Debug "check_up_to_date_e(%Ld): starting" inode;
	  Plasma_util.rpc_engine 
	    ic Ic.V1.is_up_to_date_seqno'async (inode,seqno)
  ) ++ (fun flag ->
	  match flag with
	    | true ->
		Netlog.logf `Debug "check_up_to_date_e(%Ld): true" inode;
		eps_e (`Done true) c.esys
	    | false ->
		lt # get_tid_e() ++
		  (fun tid ->
		     get_ii_serialized_e c mc tid inode true) ++
		  (fun r ->
		     let flag =
		       match r with 
			 | `ok ii -> ii.seqno = seqno
			 | _ -> false in
		     Netlog.logf `Debug "check_up_to_date_e(%Ld): %B" 
		       inode flag;
		     eps_e (`Done flag) c.esys
		  )
       )


let n_fill = 32

let n_evict = 1024L


let into_blocklist_cache c inode seqno committed index number bl =
  try
    let ce = Hashtbl.find c.inodecache inode in
    if not committed then 
      raise Not_found;
    if ce.cache_ii.seqno < seqno then (
      invalidate_ii c inode;
      raise Not_found;
    );
    if ce.cache_ii.seqno > seqno then 
      raise Not_found;
    ic_touch c inode;
    (** Garbage-collect cache: *)
    let to_remove = ref [] in
    Hashtbl.iter
      (fun i _ ->
	 if Int64.abs(Int64.sub index i) > n_evict then
	   to_remove := i :: !to_remove
      )
      ce.cache_blocks;
    List.iter
      (fun i -> Hashtbl.remove ce.cache_blocks i)
      !to_remove;
    (** Put this result into the cache *)
    (* Careful: we also want to represent empty block lists *)
    for k = 0 to number-1 do
      let i = Int64.add index (Int64.of_int k) in
      Hashtbl.replace ce.cache_blocks i []
    done;
    Array.iter
      (fun block ->
	 try
	   let l = Hashtbl.find ce.cache_blocks block.index in
	   Hashtbl.replace 
	     ce.cache_blocks block.index (block::l)
	 with Not_found -> 
	   Netlog.logf `Warning
	     "get_blocks_e: inode=%Ld got non-requested block"
	     inode
      )
      bl;
    Netlog.logf `Debug
      "get_blocks_e: inode=%Ld caching" inode
  with
    | Not_found ->
	Netlog.logf `Debug
	  "get_blocks_e: inode=%Ld not caching" inode;
	()
	  

let get_blocks_e c mc tid inode index number =
  let numberL = Int64.of_int number in
  Netlog.logf `Debug
    "get_blocks_e: inode=%Ld index=%Ld number=%d" inode index number;
  Plasma_util.rpc_engine
    mc Fsys.V1.get_blocks'async(tid,inode,index,numberL)
  ++ (fun r ->
	( match r with
	    | `ok bl ->
		Netlog.logf `Debug
		  "get_blocks_e: inode=%Ld got %d blocks" 
		  inode (Array.length bl);
		(* If we did not get _any_ blocks, we don't know
		   seqno, and cannot do the up-to-date checks. 
		   It is now worthwhile to query the seqno alone,
		   so we can do these checks.
		 *)
		if bl=[| |] then (
		  (* So get ii. It is possible that this is even a newer
		     version than at the time of the get_blocks execution.
		     No big problem, but we have to check for this.
		   *)
		  try
		    let ce = Hashtbl.find c.inodecache inode in
		    let old_seqno = ce.cache_ii.seqno in
		    get_ii_e c mc tid inode
		    ++ (fun ir ->
			  ( match ir with
			      | `ok ii ->
				  if ii.seqno = old_seqno then
				    into_blocklist_cache 
				      c inode ii.seqno ii.committed index
				      number bl
				  else
				    Netlog.logf `Debug
				      "get_blocks_e: inode=%Ld outdated"  inode
			      | _ -> ()
			  );
			  eps_e (`Done r) c.esys
		       )
		  with Not_found ->
		    (* No ce. We cannot put into the cache anyway *)
		    Netlog.logf `Debug
		      "get_blocks_e: inode=%Ld missing cache"  inode;
		    eps_e (`Done r) c.esys
		)
		else (
		  (* Standard case: get seqno/committed from a blocklist
		     element (these fields are identical for all elements)
		   *)
		  let bl0 = bl.(0) in
		  into_blocklist_cache 
		    c inode bl0.inode_seqno bl0.inode_committed index number bl;
		  eps_e (`Done r) c.esys
		)
	    | _ ->
		eps_e (`Done r) c.esys
	)
     )
	  

let get_blocks_sticky_e c mc lt inode index number =
  (* remember the get_blocks request *)
  let is_sticky = not (Hashtbl.mem c.inodecache_getting_blocks inode) in
  if is_sticky then (
    let e1, signal = Uq_engines.signal_engine c.esys in
    Hashtbl.replace c.inodecache_getting_blocks inode (index,number,e1);
    let e = 
      lt # get_tid_e() ++
	(fun tid -> get_blocks_e c mc tid inode index number) in
    let fin _ = 
      Hashtbl.remove c.inodecache_getting_blocks inode;
      signal (`Done ()) in
    Uq_engines.when_state
      ~is_done:fin ~is_error:fin ~is_aborted:fin e;
    e
  )
  else
    lt # get_tid_e() ++
      (fun tid -> get_blocks_e c mc tid inode index number)


let get_blocks_and_fill_e c mc lt inode index =
  (** Fill the cache with some blocks, but return only [index].

      Usually, clients iterate over the blocks in either ascending or
      descending order. We try to adapt to this:
       - if [index-1] is in the cache, but [index] is missing: conclude
         ascending order, and fill the range [index..index+n] for some [n]
       - if [index+1] is in the cache, but [index] is missing: conclude
         descending order, and fill the range [index-n..index] for some [n]
       - if both [index-1] and [index+1] are in the cache, or both
         are missing, or [index] is already loaded: assume unspecific
         order, and fill the range [index-n/2..index+n/2]

      We also remove entries from the cache: All blocks farer away from index
      than n_evict are deleted (in get_block_e).
   *)
  let ce_lz = lazy(Hashtbl.find c.inodecache inode) in
  let in_cache i =
    try 
      let ce = Lazy.force ce_lz in
      Hashtbl.mem ce.cache_blocks i
    with Not_found -> false in
  let index_in_cache = in_cache index in
  let pred_in_cache = in_cache (Int64.pred index) in
  let succ_in_cache = in_cache (Int64.succ index) in
  let ascending = pred_in_cache && not succ_in_cache && not index_in_cache in
  let descending = not pred_in_cache && succ_in_cache && not index_in_cache in
  (* let other = not ascending && not descending in *)

  let n_fillL = Int64.of_int n_fill in
  let min_index, n =
    if c.cacheclient = None then  (* cache is off *)
      index, 1
    else if ascending then 
      index, n_fill
    else if descending then
      max 0L (Int64.succ(Int64.sub index n_fillL)), n_fill
    else
      max 0L (Int64.succ(Int64.sub index (Int64.div n_fillL 2L))), n_fill in

  get_blocks_sticky_e c mc lt inode min_index n
  ++ (fun r ->
	match r with
	  | `ok bl ->
	      (** Extract only the requested index: *)
	      let bl' =
		Array.of_list
		  (List.fold_left
		     (fun acc block ->
			if block.index = index then
			  block :: acc
			else
			  acc
		     )
		     []
		     (Array.to_list bl)
		  ) in
	      eps_e (`Done (`ok bl')) c.esys
	  | #Plasma_util.errno as errno ->
	      eps_e (`Done errno) c.esys
     )

let get_blocks_from_cache_e c inode index =
  get_ii_from_cache_e c inode
  ++ (fun ii_opt ->
	try
	  match ii_opt with
	    | None -> raise Not_found
	    | Some _ ->
		let ce = Hashtbl.find c.inodecache inode in
		ic_touch c inode;
		let blocks = 
		  Array.of_list(Hashtbl.find ce.cache_blocks index) in
		eps_e (`Done (Some blocks)) c.esys
	with
	  | Not_found ->
	      eps_e (`Done None) c.esys
     )


let get_blocks_from_cache_with_ii_refresh_e c mc lt inode index =
  get_ii_via_cache_e c mc lt inode
  ++ (fun r ->
	match r with
	  | `ok ii ->
	      ( try
		  let ce = Hashtbl.find c.inodecache inode in
		  ic_touch c inode;
		  let blocks = 
		    Array.of_list(Hashtbl.find ce.cache_blocks index) in
		  eps_e (`Done (Some (`ok blocks))) c.esys
		with
		  | Not_found ->
		      eps_e (`Done None) c.esys
	      )
	  | #Plasma_util.errno as errno ->
	      eps_e (`Done (Some errno)) c.esys
     )


let rec get_blocks_via_cache_e c mc lt inode index =
  if c.cacheclient = None then (
    (* no cache available: get directly *)
    Netlog.logf `Debug
      "get_blocks_via_cache_e inode=%Ld index=%Ld: no cache"
      inode index;
    lt # get_tid_e()
    ++ (fun tid ->
	  get_blocks_e c mc tid inode index 1
       )
  ) else (
    get_blocks_from_cache_with_ii_refresh_e c mc lt inode index
    ++ (fun blocks_opt ->
	  match blocks_opt with
	    | Some (`ok blocks) ->
		Netlog.logf `Debug
		  "get_blocks_via_cache inode=%Ld index=%Ld: from cache"
		  inode index;
		eps_e (`Done (`ok blocks)) c.esys
	    | _ ->
		(* Get the data from the server. There might already be
		   a running get_blocks query
		 *)
		let reaction =
		  try 
		    let (i,l,e) =
		      Hashtbl.find c.inodecache_getting_blocks inode in
		    if index >= i && Int64.sub index i < Int64.of_int l then
		      `Wait_for_request e
		    else
		      `Request_single_block
		  with Not_found ->
		    `Request_several_blocks in
		match reaction with
		  | `Request_single_block ->
		      Netlog.logf `Debug
			"get_blocks_via_cache inode=%Ld index=%Ld: single block"
			inode index;
		      lt # get_tid_e()
		      ++ (fun tid -> 
			    get_blocks_e c mc tid inode index 1)
		  | `Request_several_blocks ->
		      Netlog.logf `Debug
			"get_blocks_via_cache inode=%Ld index=%Ld: \
                         several blocks"
			inode index;
		      get_blocks_and_fill_e c mc lt inode index
		  | `Wait_for_request e ->
		      (* Wait until the current request is done, and
			 then check the cache again.
		       *)
		      Netlog.logf `Debug
			"get_blocks_from_cache_e inode=%Ld index=%Ld: waiting"
			inode index;
		      e
		      ++ (fun () ->
			    get_blocks_via_cache_e c mc lt inode index
			 )
       )
  )
